/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/base.hpp

    Copyright © 2014, 2016 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    ---------------------------------------------------------------------- copyright and license ---
*/


#ifndef _BASE_HPP_
#define _BASE_HPP_


#include "config.h"

#include <exception>

#include "string.hpp"
#include "ui.hpp"


/**
    Base class for all other classes.

    It provides debugging facilities.
*/
class object_t {

    #if defined( DEBUG )

        public :

            explicit object_t( char const * type );
            virtual ~object_t();

            object_t const *    _self_()  const;
            unsigned int        _no_()    const;
            string_t            _id_()    const;
            virtual string_t    _str_()   const;

        private :

            char const * const  _type;
            unsigned int const  _no;
            string_t const      _id;
    #else

        public :

            explicit object_t( char const * type ) {};

    #endif

}; // object_t


#if defined( DEBUG )
    object_t const * _self_();
    std::ostream & operator <<( std::ostream & stream, object_t const & obj );
#endif

#if __cplusplus < 201103L
    #define DELETE
#else
    #define DELETE = delete
#endif
/**
    Base class for non-copyable objects.

    It hides copy constructor and assignment operator, so making copies of objects it not possible.
*/
class non_copyable_object_t : public object_t {
    public :
        explicit non_copyable_object_t( char const * type ) : object_t( type ) {};
        non_copyable_object_t( non_copyable_object_t const & ) DELETE;
        non_copyable_object_t & operator =( non_copyable_object_t const & ) DELETE;
}; // non_copyable_object_t


#if defined( DEBUG )
    void _dbg( object_t const * obj, char const * file, int line, string_t const & mesg );
    #define DBG1( _OBJ_, _ARGS_ ) _dbg( _OBJ_, __FILE__, __LINE__, STR( _ARGS_ ) )
    #define WRN1( _OBJ_, _ARGS_ ) {                                                                \
        string_t _MSG_ = STR( _ARGS_ );                                                            \
        DBG1( _OBJ_, "Issuing warning: " << _MSG_ );                                               \
        msg( sl_warning, _MSG_ );                                                                  \
    }
    #define ERR1( _OBJ_, _ARGS_, ... ) {                                                           \
        string_t _PFX_ = STR( ( _OBJ_ == NULL ? "" : _OBJ_->_str_() + ": " ) );                    \
        string_t _MSG_ = STR( _ARGS_ );                                                            \
        DBG1( _OBJ_, "Throwing exception: " << _MSG_ );                                            \
        throw error_t( _PFX_ + _MSG_, ## __VA_ARGS__ );                                            \
    }
#else
    #define DBG1( _OBJ_, _ARGS_ )
    #define WRN1( _OBJ_, _ARGS_ ) {                                                                \
        string_t _MSG_ = STR( _ARGS_ );                                                            \
        msg( sl_warning, _MSG_ );                                                                  \
    }
    #define ERR1( _OBJ_, _ARGS_, ... ) {                                                           \
        string_t _MSG_ = STR( _ARGS_ );                                                            \
        throw error_t( _MSG_, ## __VA_ARGS__ );                                                    \
    }
#endif

#define DBG( _ARGS_ )       DBG1( _self_(), _ARGS_ )
#define INF( _ARGS_ )       msg( sl_info, STR( _ARGS_ ) )
#define WRN( _ARGS_ )       WRN1( _self_(), _ARGS_ )
#define ERR( _ARGS_, ... )  ERR1( _self_(), _ARGS_, ## __VA_ARGS__ )

#define CATCH_ALL1( _OBJ_, _STATEMENT_ ) {                                                         \
    try {                                                                                          \
        _STATEMENT_;                                                                               \
    } catch ( std::exception const & ex ) {                                                        \
        WRN1( _OBJ_, ex.what() );                                                                  \
    } catch ( ... ) {                                                                              \
        WRN1( _OBJ_, "Unknown exception occurred" );                                               \
    };                                                                                             \
}

#define CATCH_ALL( _STATEMENT_ ) \
    CATCH_ALL1( _self_(), _STATEMENT_ )

#if defined( DEBUG )

    class _func_tracer_t {
        public :
            _func_tracer_t( object_t const * obj, char const * file, int line, char const * func );
            ~_func_tracer_t();
        private :
            object_t const * const  _obj;
            char const * const      _file;
            int const               _line;
            char const * const      _func;
    }; // class _func_tracer_t

    #define TRACE1( _OBJ_ ) \
        _func_tracer_t _tracer_ ## __COUNTER__( _OBJ_, __FILE__, __LINE__, __FUNCTION__ )
    #define TRACE() \
        TRACE1( _self_() )
#else

    #define TRACE1( _OBJ_ )
    #define TRACE()

#endif

#endif // _OBJECT_HPP_

// end of file //
