/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/main.cpp

    Copyright © 2014, 2016 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    ---------------------------------------------------------------------- copyright and license ---
*/


#include "config.h"
#include "base.hpp"
#include "switcher.hpp"
#include "ui.hpp"

#include <exception>
#include <iostream>

#include <cstdlib>

#include <argp.h>


/*
    TODO:

        *   Detect changes in runtime (autorepeat rate changed, keyboard changed, etc)
        *   Two keyboards?

*/


/**
    Program version string.

    It is printed in respond to `--version` option. The string may include copyright.
*/
char const * argp_program_version =
    PACKAGE_STRING "\n"
    "Copyright (C) " PACKAGE_COPYRIGHT_YEAR " " PACKAGE_COPYRIGHT_HOLDER ".\n"
    "License " PACKAGE_LICENSE " <" PACKAGE_LICENSE_URL ">.\n"
    "This is free software: you are free to change and redistribute it." "\n"
    "There is NO WARRANTY, to the extent permitted by law."
    ;

/**
    Email address to send bug reports to.

    May be `NULL`.
*/
char const * argp_program_bug_address = PACKAGE_BUGREPORT;


error_t argp_err_exit_status = 2;


void msg( severity_level_t level, string_t const & message ) {
    string_t prefix = "?";
    switch ( level ) {
        case sl_error : {
            prefix = "E";
        } break;
        case sl_warning : {
            prefix = "W";
        } break;
        case sl_info : {
            prefix = "";
        } break;
        case sl_debug : {
            prefix = "#";
        } break;
    };
    std::cerr << STR(
        ( prefix.empty() ? "" : STR( program_invocation_short_name << ": " << prefix << " " ) )
        << message << "\n"
    );
}; // msg


class app_t : public non_copyable_object_t {

    public :

        class error_t : public std::runtime_error {
            public :
                explicit error_t( string_t const & what ) : std::runtime_error( what ) {};
        }; // class error_t

    public :


        app_t(
            int     argc,
            char *  argv[]
        ) :
            non_copyable_object_t( __FUNCTION__ ),
            prog_name( program_invocation_short_name ),
            bell( false )
        {
            TRACE();
            posix::signal::set_t set;
            set += SIGINT;
            set += SIGTERM;
            posix::signal::mask( SIG_BLOCK, set );
            parse_argv( argc, argv );
            INF( PACKAGE_STRING );
            INF(
                PACKAGE_NAME " is a keyboard layout selector for X Window System: "
                "it selects specified keyboard layout when corresponding key is tapped."
            );
            init_x();
            layout_switcher_t switcher( * kb, map, bell );
            switcher.start();
            strings_t const & names = key_names();
            for ( map_t::const_iterator it = map.begin(); it != map.end(); ++ it ) {
                INF(
                    "Tap `" << names[ it->first ] << "' "
                    << "to select layout `" << it->second + 1 << "'."
                );
            }; // for it
            INF( "Press `Ctrl'+`C' to exit..." );
            posix::signal::wait( set );
            switcher.stop();
            INF( "Done." );
        }; // ctor


    private :

        typedef x::kb_t::desc_t::key_code_range_t key_code_range_t;
        typedef layout_switcher_t::map_t          map_t;


        enum opt_key_t {

            ok_list_keys    = 1000,
            ok_list_layouts,
            ok_bell,

        }; // enum opt_key_t


        static
        ::error_t
        parse_opt_or_arg(
            int                 key,        ///< Option key (a short name or integer id).
            char *              _arg,       ///< String to parse.
            struct argp_state * state
        ) {

            char const * arg = _arg;
                //  It is a little trick. We are not going to change argument, but parameter
                //  should be of `char *` type.

            app_t * app = reinterpret_cast< app_t * >( state->input );

            try {

                switch ( key ) {

                    case ok_list_keys : {
                        app->list_keys();
                        exit( EXIT_SUCCESS );
                    } break;

                    case ok_list_layouts : {
                        app->list_layouts();
                        exit( EXIT_SUCCESS );
                    } break;

                    case ok_bell : {
                        app->bell = true;
                    } break;

                    case ARGP_KEY_NO_ARGS : {
                        arg = "LFSH=1";
                        app->parse_arg( arg );
                        arg = "RTSH=2";
                        app->parse_arg( arg );
                    } break;

                    case ARGP_KEY_ARG : {
                        app->parse_arg( arg );
                    } break;

                    case ARGP_KEY_END : {
                    } break;

                    default : {
                        return ARGP_ERR_UNKNOWN;
                    } break;

                }; // switch

            } catch ( std::exception const & ex ) {

                char const * format = NULL;
                switch ( key ) {
                    case ARGP_KEY_ARG : {
                        format = "Bad argument `%s': %s";
                    } break;
                    case ARGP_KEY_NO_ARGS : {
                        format =
                            "Oops, bad default argument `%s': %s; "
                            "try to specify explicit argument(s) to avoid this problem";
                    } break;
                    default : {
                        format = "Bad option `%s': %s";
                        arg = state->argv[ state->next - 1 ];
                    } break;
                }; // switch
                argp_error( state, format, arg, ex.what() );

            }; // try

            return 0;

        }; // parse_opt_or_arg


        void
        parse_arg(
            string_t const & arg
        ) {
            string_pair_t parts    = split( '=', arg );
            string_t key           = parts.first;
            string_t layout        = parts.second;
            ssize_t  key_code      = -1;
            string_t key_name      = "";
            ssize_t  layout_number = -1;
            string_t layout_name   = "";
            if ( key.empty() ) {
                ERR( "No key specified" );
            }; // if
            if ( natural_number( key ) ) { // Looks like a key code.
                try {
                    key_code = val< ssize_t >( key );
                } catch ( std::exception const & ex ) {
                    ERR( "Bad key code `" << key << "': " << ex.what() );
                }; // try
                key_code_range_t range = key_code_range();
                if ( not ( range.first <= key_code and key_code <= range.second ) ) {
                    ERR(
                        "Bad key code `" << key << "': "
                        << "Out of range; Key code should be in range [" << range.first << " .. "
                        << range.second << "]"
                    );
                }; // if
            } else {                       // Looks like a key name.
                key_name = key;
                if ( left( key_name, 1 ) == "<" and right( key_name, 1 ) == ">" ) {
                    key_name = key_name.substr( 1, key_name.size() - 2 );
                }; // if
                key_code = find( key_name, key_names() );
                if ( key_code < 0 ) {
                    // If not found, try to find upper-case variant of key name.
                    key_code = find( uc( key_name ), key_names() );
                }; // if
                if ( key_code < 0 ) {
                    ERR( "Unknown key name `" << key_name << "'" );
                }; // if
            }; // if
            // Check if the key is already assigned.
            map_t::const_iterator it = map.find( key_code );
            if ( it != map.end() ) {
                ERR(
                    "Key `" << key_code << ( key_name.empty() ? "" : ":" + key_name ) << "' "
                    << "is already assigned"
                );
            }; // if
            if ( layout.empty() ) {
                ERR( "No layout specified" );
            }; // if
            if ( natural_number( layout ) ) {
                try {
                    layout_number = val< ssize_t >( layout );
                } catch ( std::exception const & ex ) {
                    ERR( "Bad layout number `" << layout << "': " << ex.what() );
                }; // try
                // Incapsulation broken. TODO: Fix it.
                if ( layout_number > 4 ) {
                    ERR(
                        "Bad layout number `" << layout_name << "': "
                        << "Too big; Maximal layout number is 4."
                    );
                }; // if
            } else {
                layout_name             = layout;
                strings_t const & names = layout_names();
                layout_number           = find( layout_name, names ) + 1;
                if ( layout_number <= 0 ) {
                    string_t layouts = "";
                    for ( size_t i = 0; i < names.size(); ++ i ) {
                        if ( not names[ i ].empty() ) {
                            layouts += string_t( layouts.empty() ? "" : ", " ) + "`" + names[ i ] + "'";
                        }; // if
                    };
                    ERR(
                        "Unknown layout name `" << layout_name << "'; "
                        << " known layouts are: " << layouts
                    );
                }; // if
            }; // if
            map[ key_code ] = layout_number - 1;
            DBG(
                key_code << ":" << key_names()[ key_code ]
                << "="
                << layout_number << ( layout_name.empty() ? "" : ":" + layout_name )
            );
        }; // parse_arg


        void
        parse_argv(
            int     argc,
            char *  argv[]
        ) {

            argp_option opts[] = {

                { "Other options:", 0,                  NULL, OPTION_DOC | OPTION_NO_USAGE,
                    "",
                    200 },
                { "bell",           ok_bell,            NULL, 0,
                    "Ring a bell when a layout is selected",
                    201 },

                { "Help options:",  0,                  NULL, OPTION_DOC | OPTION_NO_USAGE,
                    "",
                    -2  },
                { "list-keys",      ok_list_keys,       NULL, 0,
                    "List key codes and names",
                    -1  },
                { "list-layouts",   ok_list_layouts,    NULL, 0,
                    "List layout numbers and names",
                    -1  },

                { 0 }

            }; // opts

            string_t doc = STR(
                "\n"
                PACKAGE_NAME " is a keyboard layout selector for X Window System: "
                "it selects specified keyboard LAYOUT when corresponding KEY is tapped."
                "\v"    // Options will be listed here.
                "KEY is either a key code (unsigned integer) "
                "or a symbolic key name (`LFSH', `RALT', etc). "
                "Run `" << prog_name << " --list-keys' "
                "to see the list of keys currently known to the X Window server. "
                "\n\n"
                "\"Tap\" means that a single key is pressed and quickly released. "
                "If a key is released slowly, "
                "or pressed/released in combination with other key(s), "
                "it is not a tap. "
                "\n\n"
                "LAYOUT is an Xkb \"group\", "
                "either a group number (from 1 to 4) or its name. "
                "Run `" << prog_name << " --list-layouts' "
                "to see the list of currently configured layouts. "
                "\n\n"
                "If no arguments specified, `LFSH=1' and `RTSH=2' are used by default."
                "\n\n"
                "Exit status:\n"
                "  0 -- Everything is ok.\n"
                "  " << argp_err_exit_status << " -- Usage (command-line) error.\n"
                "  255 -- Runtime error.\n"
                "\n"
                "For more detaled documenatation, run `man " PACKAGE "'.\n"
                PACKAGE_NAME " home page: " PACKAGE_URL "."
            );

            argp parser = { opts, parse_opt_or_arg, "[KEY=LAYOUT...]", doc.c_str() };
            ::error_t error = argp_parse( & parser, argc, argv, 0, 0, this );
            if ( error ) {
                exit( argp_err_exit_status );
            }; // if

        }; // parse_argv


        void
        init_x(
        ) {
            if ( display.get() == NULL ) {
                x::init();
                display.reset( new x::display_t );
                kb.reset( new x::kb_t( * display ) );
                x::kb_t::desc_p kb_desc( kb->desc() );
                _key_code_range = kb_desc->key_code_range();
                _key_names      = kb_desc->key_names();
                _layout_names   = kb_desc->group_names();
            }; // if
        }; // init_x


        void
        list_keys(
        ) {
            strings_t const & names = key_names();
            for ( size_t i = 0; i < names.size(); ++ i ) {
                if ( not names[ i ].empty() ) {
                    std::cout << i << ":" << names[ i ] << "\n";
                }; // if
            }; // for i
        }; // list_keys


        void
        list_layouts(
        ) {
            strings_t const & names = layout_names();
            for ( size_t i = 0; i < names.size(); ++ i ) {
                if ( not names[ i ].empty() ) {
                    std::cout
                        << i + 1 << ":"
                        << names[ i ]
                        << "\n";
                }; // if
            }; // for i
        }; // list_layouts


        key_code_range_t const &
        key_code_range() {
            init_x();
            return _key_code_range;
        }; // key_code_range


        strings_t const &
        key_names() {
            init_x();
            return _key_names;
        }; // key_names


        strings_t const &
        layout_names() {
            init_x();
            return _layout_names;
        }; // layout_names


    private :

        char const * const  prog_name;

        x::display_p        display;
        x::kb_p             kb;

        key_code_range_t    _key_code_range;
        strings_t           _key_names;
        strings_t           _layout_names;

        map_t               map;
        bool                bell;

}; // class app_t


int
main(
    int     argc,
    char *  argv[]
) {
    int status = 0;
    try {
        app_t app( argc, argv );
    } catch ( std::exception const & ex ) {
        msg( sl_error, ex.what() );
        status = 1;
    } catch ( ... ) {
        msg( sl_error, "Unknown exception occurred" );
        status = 2;
    }; // try
    return status;
}; // main


// end of file //
