/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/tapper.cpp

    Copyright © 2014, 2016 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    ---------------------------------------------------------------------- copyright and license ---
*/


#include "tapper.hpp"

#include <limits>

#include <X11/Xproto.h>     // xEvent


// -------------------------------------------------------------------------------------------------
// tapper_t
// -------------------------------------------------------------------------------------------------


tapper_t::tapper_t(
    x::kb_t & kb
) :
    non_copyable_object_t( __FUNCTION__ ),
    _kb( kb ),
    interceptor( * this )
{
    autorepeat_rate = _kb.autorepeat_rate();
}; // ctor


tapper_t::~tapper_t(
) {
}; // dtor


void
tapper_t::start(
) {
    TRACE();
    interceptor.enable();
}; // start


void
tapper_t::stop(
) {
    TRACE();
    interceptor.disable();
}; // stop


tapper_t::interceptor_t::interceptor_t(
    tapper_t &  _master
) :
    master( _master ),
    state( std::numeric_limits< KeyCode >::max() ),
    pressed_keys( 0 ),
    last_pressed_key( 0 ),
    pressed_at( 0 )
{
}; // ctor


void
tapper_t::interceptor_t::on_intercept(
    XRecordInterceptData * data
) {
    TRACE();
    if ( data->category == XRecordFromServer ) {
        /*
            Decoding data described in:
            http://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.html#events:input
        */
        if ( data->data_len * 4 < sizeof( xEvent ) ) {
            // `data_len` is size of data in 4-byte units.
            //~ ERR( "Oops" );
        }; // if
        xEvent * event = reinterpret_cast< xEvent * >( data->data );
        // My experiments show that data->server_time == event->u.keyButtonPointer.time.
        switch ( event->u.u.type ) {
            case KeyPress   :
            case KeyRelease : {
                bool    pressed  = ( event->u.u.type == KeyPress );
                int     key_code = event->u.u.detail;
                Time    time     = event->u.keyButtonPointer.time;
                //~ int     state    = event->u.keyButtonPointer.state;
                if ( pressed ) {
                    if ( state[ key_code ] ) {
                        // Look like key press is autorepeating.
                        if ( last_pressed_key != 0 ) {
                            // BTW, we are having autorepeat timeout for free...
                            //~ master.autorepeat_rate.timeout = time - pressed_at;
                        }; // if
                        /*
                            No need in updating keyboard state -- we already know the key is
                            pressed. Number of pressed keys is not changed too. But let us reset
                            the last pressed key code and time -- if autorepeating takes place
                            this is not a "tap".
                        */
                        last_pressed_key = 0;
                        pressed_at       = 0;
                        DBG( key_code << "↡ " << pressed_keys );
                    } else {
                        // Update keyboard state.
                        state[ key_code ] = true;
                        pressed_keys += 1;
                        // And remember the last pressed key code and time when it was pressed.
                        last_pressed_key = key_code;
                        pressed_at       = time;
                        DBG( "k" << key_code << "↓ " << pressed_keys );
                    }; // if
                } else {
                    /*
                        This check is a must because unconditional decreasing `pressed_keys` may
                        cause wrong result. For example, if the program started from the command
                        line, the first received event will likely be releasing of Enter key.
                    */
                    if ( state[ key_code ] ) {
                        state[ key_code ] = false;
                        pressed_keys -= 1;
                    }; // if
                    if (
                        key_code == last_pressed_key
                        and pressed_keys == 0
                        and time - pressed_at <= master.autorepeat_rate.timeout
                    ) {
                        master.on_tap( key_code );
                    }; // if
                    last_pressed_key = 0;
                    pressed_at       = 0;
                    DBG( "k" << key_code << "↑ " << pressed_keys );
                }; // if
            } break;
            case ButtonPress   :
            case ButtonRelease : {
                bool    pressed  = ( event->u.u.type == ButtonPress );
                int     button   = event->u.u.detail;
                //~ Time    time     = event->u.keyButtonPointer.time;
                static_cast< void >( button );      // Avoid cppcheck warning "buton never used".
                last_pressed_key = 0;
                pressed_at       = 0;
                if ( pressed ) {
                    DBG( "b" << button << "↓" );
                } else {
                    DBG( "b" << button << "↑" );
                }; // if
            } break;
        }; // switch
    }; // if
}; // on_intercept


// end of file //
