/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/x.hpp

    Copyright © 2014, 2016 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    ---------------------------------------------------------------------- copyright and license ---
*/


#ifndef _X_HPP_
#define _X_HPP_


#include "config.h"
#include "base.hpp"
#include "posix.hpp"

#include <memory>           // std::auto_ptr/std::unique_ptr
#include <stdexcept>        // std::runtime_error

#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/record.h>
#include <X11/extensions/XKB.h>


namespace x {

    // std::auto_ptr is deprecated in C++11, let us use std::unique_ptr instead.
    #if __cplusplus >= 201103L
        template< typename type_t >
        using auto_ptr = std::unique_ptr< type_t >;
    #else
        using std::auto_ptr;
    #endif

    void init();


    // ---------------------------------------------------------------------------------------------
    // error_t
    // ---------------------------------------------------------------------------------------------

    class error_t : public std::runtime_error {
        public :
            explicit error_t( string_t const & msg );
    }; // class error_t


    // ---------------------------------------------------------------------------------------------
    // display_t
    // ---------------------------------------------------------------------------------------------

    class display_t : public non_copyable_object_t {

        public :

            explicit display_t();
            virtual ~display_t();

            int connection() const;

            void flush();
            void sync( bool discard = false );

            #if defined( DEBUG )
                virtual string_t _str_() const;
            #endif

        public :

            Display * rep;

        private :

            string_t  _name;
            int       _conn;

    }; // class display_t

    typedef auto_ptr< display_t > display_p;


    // ---------------------------------------------------------------------------------------------
    // atom_t
    // ---------------------------------------------------------------------------------------------

    class atom_t : public object_t {

        public :

            explicit atom_t( display_t & display, Atom atom );
            explicit atom_t( Display * display, Atom atom );

            string_t name();

            #if defined( DEBUG )
                virtual string_t _str_() const;
            #endif

        private:

            Display *   _display;
            Atom        rep;
            string_t    _name;

    }; // atom_t


    namespace record {


        // -----------------------------------------------------------------------------------------
        // intercept_data_t
        // -----------------------------------------------------------------------------------------

        class intercept_data_t : public non_copyable_object_t {

            public :

                explicit intercept_data_t( XRecordInterceptData * data = NULL );
                virtual ~intercept_data_t();

            public :

                XRecordInterceptData * data;

        }; // class intercept_data_t


        // -----------------------------------------------------------------------------------------
        // range_t
        // -----------------------------------------------------------------------------------------

        class range_t : public non_copyable_object_t {

            public :

                explicit range_t( unsigned char first, unsigned char last );
                virtual ~range_t();

            public :

                XRecordRange * range;

        }; // class range_t


        // -----------------------------------------------------------------------------------------
        // context_t
        // -----------------------------------------------------------------------------------------

        class context_t : public non_copyable_object_t {

            public :

                explicit context_t(
                    int               datum_flags = 0,
                    XRecordClientSpec client      = XRecordAllClients
                );
                virtual ~context_t();

                void enable();
                void disable();

            protected :

                virtual void on_intercept( XRecordInterceptData * data ) = 0;

            private :

                static void interceptor( XPointer closure, XRecordInterceptData * data );

                class thread_t : public posix::thread_t {
                    friend class context_t;
                    public :
                        explicit thread_t( display_t & dsp );
                    protected :
                        virtual void body();
                    private :
                        display_t & _display;
                }; // thread_t

            private :

                enum state_t {
                    inited,
                    enabled,
                    disabled
                }; // enum state_t

                state_t         _state;
                XRecordContext  _handle;
                display_t       _ctrl;
                display_t       _data;
                thread_t        _thread;

        }; // class context_t


    }; // namespace record


    // ---------------------------------------------------------------------------------------------
    // kb_t
    // ---------------------------------------------------------------------------------------------

    class kb_t : public non_copyable_object_t {

        public :

            struct autorepeat_rate_t {
                autorepeat_rate_t();
                unsigned int timeout;
                unsigned int interval;
            }; // struct autorepeat_rate_t

            class desc_t : public non_copyable_object_t {

                friend class kb_t;

                private :

                    explicit desc_t( kb_t & kb, unsigned int which );

                public :

                    typedef std::pair< int, int > key_code_range_t;

                    virtual ~desc_t();

                    key_code_range_t key_code_range()   const;
                    strings_t        key_names()        const;
                    strings_t        group_names()      const;

                private :

                    XkbDescPtr  rep;

            }; // class desc_t

            typedef auto_ptr< desc_t > desc_p;

        public :

            explicit kb_t( display_t & display );

            display_t &       display();
            autorepeat_rate_t autorepeat_rate();
            void              lock_group( unsigned int group );
            void              bell();
            desc_p            desc( unsigned int which = XkbAllComponentsMask );

        private :

            display_t & _display;

    }; // class kb_t

    typedef auto_ptr< kb_t >   kb_p;


}; // namespace x


#endif // _X_HPP_


// end of file //
